import os
import asyncio
import threading

from dotenv import load_dotenv
load_dotenv()

from rich.console import Console
console = Console()

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

# Internal modules
from gai.lib.diagnostics import free_mem
from gai.lib.color import red, green
from gai.lib.config import GaiConfig, GaiGeneratorConfig
from gai.lib.logging import getLogger
logger = getLogger(__name__)

def get_app_version(pyproject_toml):
    import toml
    with open(pyproject_toml) as f:
        pyproject = toml.load(f)
    return pyproject["project"]["version"]

# This tells fastapi which path to host the swagger ui page.
def get_swagger_url():
    swagger_url=None
    if "SWAGGER_URL" in os.environ and os.environ["SWAGGER_URL"]:
        swagger_url=os.environ["SWAGGER_URL"]
        logger.info(f"swagger={swagger_url}")
    else:
        logger.info("swagger_url=disabled.")
    return swagger_url

def configure_cors(app: FastAPI):
    allowed_origins_str = "*"
    if "CORS_ALLOWED" in os.environ:
        allowed_origins_str = os.environ["CORS_ALLOWED"]    # from .env
    allowed_origins = allowed_origins_str.split(",")  # split the string into a list
    logger.info(f"allow_origins={allowed_origins}")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=allowed_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

def configure_semaphore():
    use_semaphore = os.getenv("USE_SEMAPHORE", "False").lower() == "true"
    semaphore = None
    if use_semaphore:
        logger.info("Using semaphore")
        import asyncio
        semaphore = asyncio.Semaphore(2)
    return semaphore

async def acquire_semaphore(semaphore):
    while semaphore:
        try:
            await asyncio.wait_for(semaphore.acquire(), timeout=0.1)
            break
        except asyncio.TimeoutError:
            logger.warn("_streaming: Server is busy")
            await asyncio.sleep(1)

def release_semaphore(semaphore):
    if semaphore:
        semaphore.release()
        logger.debug("_streaming: Server is available")

def get_startup_event(app, category:str, pyproject_toml, generator_config:GaiGeneratorConfig):

    async def startup_event():
        
        try:
            # check freemem before loading the model
            free_mem()

            # version check
            logger.info(f"Starting Gai LLM Service ({category}) {get_app_version(pyproject_toml)}")
            
            # extract the default generator config for a category and add it to the app state

            app.state.generator_config = generator_config

            # initialize host and add it to the app state
            host = SingletonHost.GetInstanceFromConfig(generator_config)
            host.load()
            logger.info(f"Model loaded = [{generator_config.name}]")
            app.state.host = host

            # check freemem after loading the model
            free_mem()    
        except Exception as e:
            logger.error(f"Failed to load default model: {e}")
            raise e

        app.state.host = host

    return startup_event

def get_shutdown_event(app):
    
    async def shutdown_event():
        host = app.state.host
        if host:
            host.unload()

    return shutdown_event

def create_app(pyproject_toml, category, generator_config):

    app=FastAPI(
        title="Gai Generators Service",
        description="""Gai Generators Service""",
        version=get_app_version(pyproject_toml),
        docs_url=get_swagger_url()
        )
    configure_cors(app)

    # Event Handlers
    app.add_event_handler("startup", get_startup_event(app, category=category, pyproject_toml=pyproject_toml, generator_config=generator_config))
    app.add_event_handler("shutdown", get_shutdown_event(app))

    return app

class SingletonHost:
    __instance = None       # singleton

    @staticmethod
    def GetInstanceFromPath(generator_name,config_path=None,verbose=True):
        """Static method to access this singleton class's instance."""
        config_path=os.path.expanduser(config_path)
        gai_config = GaiConfig.from_path(config_path)
        gen_config = gai_config.generators[generator_name]
        
        if SingletonHost.__instance == None:
            SingletonHost.__instance=SingletonHost(gen_config,verbose=verbose)
        else:
            # Override __instance's config and verbose if it already exists
            SingletonHost.__instance.config=gen_config
            SingletonHost.__instance.__verbose=verbose
        return SingletonHost.__instance

    @staticmethod
    def GetInstanceFromConfig(config:GaiGeneratorConfig,verbose=True):
        """Static method to access this singleton class's instance."""
        if SingletonHost.__instance == None:
            SingletonHost.__instance=SingletonHost(config,verbose=verbose)
        else:
            # Override __instance's config and verbose if it already exists
            SingletonHost.__instance.config=config
            SingletonHost.__instance.__verbose=verbose
        return SingletonHost.__instance

    def __init__(self,config:GaiGeneratorConfig,verbose=True):
        self.__verbose=verbose

        """Virtually private constructor."""
        if SingletonHost.__instance is not None:
            raise Exception(
                "SingletonHost: This class is a singleton! Access using GetInstance().")
        else:
            # This class only has 4 attributes

            # config is always the first to be loaded from constructor
            self.config = config

            # generator is loaded by calling load()
            self.generator = None

            # generator_name matches config["generator_name"] and is loaded only when self.generator is successfully loaded 
            self.generator_name = None

            # Finally, for thread safety and since this is run locally, use semaphore to ensure only one thread can access the generator at a time
            self.semaphore = threading.Semaphore(1)

            SingletonHost.__instance = self

    def __enter__(self):
        self.load()
        return self
    
    def __exit__(self,exc_type, exc_value,traceback):
        self.unload()
        import gc,torch
        gc.collect()
        torch.cuda.empty_cache()

    # This is idempotent
    def load(self):

        if self.generator_name == self.config.name:
            logger.debug(
                "SingletonHost.load: Generator is already loaded. Skip loading.")
            return self

        if self.generator_name and self.generator_name != self.config.name:
            logger.debug(
                "SingletonHost.load: New generator_name specified, unload current generator.")
            if self.generator:
                self.unload()
            return self

        try:
            target_name=self.config.name
            logger.info(f"SingletonHost: Loading generator {target_name}...")

            # Load generator using reflection
            import importlib
            module = importlib.import_module(self.config.module.name)
            class_ = getattr(module, self.config.module.class_)
            self.generator = class_(generator_config=self.config, verbose=self.__verbose)
            self.generator.load()
            self.generator_name=target_name
            return self
        except Exception as e:
            self.unload()
            logger.error(
                f"SingletonHost: Error loading generator {self.generator_name}: {e}")
            raise e
            
    def unload(self):
        if self.generator is not None:
            self.generator.unload()
            del self.generator
            self.generator = None
            self.generator_name = None
            import gc,torch
            gc.collect()
            torch.cuda.empty_cache()
        return self



import os
import yaml
from pydantic_settings import BaseSettings
from pydantic import Field
from typing import Optional, Dict
from gai.lib.utils import get_app_path
from gai.lib.logging import getLogger
logger = getLogger(__name__)

class LogConfig(BaseSettings):
    level: str = "INFO"
    format: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    datefmt: str = "%Y-%m-%d %H:%M:%S"
    filename: Optional[str] = None
    filemode: str = "a"
    stream: str = "stdout"
    loggers: Optional[Dict] = None

### GaiClientConfig

class GaiClientConfig(BaseSettings):
    client_type: str
    type: Optional[str] = None
    engine: Optional[str] = None
    model: Optional[str] = None
    name: Optional[str] = None
    url: Optional[str] = None
    env: Optional[Dict] = None
    extra: Optional[Dict] = None
    hyperparameters: Optional[Dict] = {}

    @classmethod
    def from_name(cls,name:str, file_path:str=None) -> "GaiClientConfig":
        return cls._get_client_config(name=name, file_path=file_path)

    @classmethod
    def from_dict(cls, config:dict) -> "GaiClientConfig":
        """
        Class method to create an instance of GaiClientConfig from a dictionary.

        Parameters:
            config (dict): A dictionary containing the configuration data.

        Returns:
            GaiClientConfig: An instance of GaiClientConfig populated with the configuration data.

        Usage example:
            config = GaiClientConfig.from_dict(config={
                "url": "https://api.openai.com/v1/engines/davinci-codex/completions",
                "type": "openai",
                "engine": "davinci-codex",
                "model": "davinci-codex",
                "name": "OpenAI Codex",
                "client_type": "openai"
            })
        """        
        return cls._get_client_config(config=config)

    @classmethod
    def _get_gai_config(cls, file_path:str) -> Dict:
        gai_dict = None
        try:
            with open(file_path, 'r') as f:
                gai_dict = yaml.load(f, Loader=yaml.FullLoader)
        except Exception as e:
            raise ValueError(f"GaiClientConfig: Error loading client config from file: {e}")
        return gai_dict    

    @classmethod
    def _get_client_config(
            cls,
            name: Optional[str] = None,
            config: Optional[dict] = None,
            file_path: Optional[str] = None    
        ) -> "GaiClientConfig":
        """
        Retrieves a GaiClientConfig object based on the provided arguments.

        Parameters:
            name (str, optional): The name of the configuration.
            config (dict, optional): A dictionary containing the configuration data.
            file_path (str, optional): Path to the configuration file.

        Returns:
            GaiClientConfig: The configuration object based on the provided arguments.

        Raises:
            ValueError: If the arguments are invalid or required keys are missing.

        Usage examples:
            1. Using a dict:
                config = GaiClientConfig.from_dict(config={
                    "url": "https://api.openai.com/v1/engines/davinci-codex/completions",
                    "type": "openai",
                    "engine": "davinci-codex",
                    "model": "davinci-codex",
                    "name": "OpenAI Codex",
                    "client_type": "openai"
                })

            2. Get default ttt config from a specific configuration file:
                config = ClientLLMConfig.from_name(name="ttt", file_path="config.yaml")

            3. Get default ttt config from ~/.gai/gai.yml:
                config = ClientLLMConfig.from_name(name="ttt")
        """
        
        if config:
            return cls(**config)
        
        if name:
            gai_dict = None
            try:
                app_dir=get_app_path()
                global_lib_config_path = os.path.join(app_dir, 'gai.yml')
                if file_path:
                    global_lib_config_path = file_path
                gai_dict = cls._get_gai_config(global_lib_config_path)
            except Exception as e:
                raise ValueError(f"GaiClientConfig: Error loading client config from file: {e}")

            client_config = gai_dict["clients"].get(name, None)
            if not client_config:
                raise ValueError(f"GaiClientConfig: Client Config not found. name={name}")
            return cls(**client_config)
        raise ValueError("GaiClientConfig: Invalid arguments. Either 'name' or 'config' must be provided.")

### GaiGeneratorConfig

class ModuleConfig(BaseSettings):
    name: str
    class_: str = Field(alias="class")  # Use 'class' as an alias for 'class_'

    class Config:
        allow_population_by_name = True  # Allow access via both 'class' and 'class_'

class GaiGeneratorConfig(BaseSettings):
    type: str
    engine: str
    model: str
    name: str
    hyperparameters: Optional[Dict] = {}
    extra: Optional[Dict] = None
    module: ModuleConfig
    class Config:
        extra = "allow"

    @classmethod
    def from_name(cls,name:str, file_path:str=None) -> "GaiGeneratorConfig":
        return cls._get_generator_config(name=name, file_path=file_path)
    
    @classmethod
    def from_dict(cls, config:dict) -> "GaiGeneratorConfig":
        return cls._get_generator_config(config=config)
    
    @staticmethod
    def _find_caller_module_path():
        import inspect
        """
        Finds the path of the module that originally called into the GaiGeneratorConfig,
        ignoring any calls that originate from 'site-packages'.
        """
        # Traverse the stack from the current call upwards
        for frame_info in inspect.stack():
            module = inspect.getmodule(frame_info.frame)
            if module and hasattr(module, '__file__'):
                file_path = module.__file__
                # Skip frames where the file path contains 'site-packages'
                if 'site-packages' not in file_path:
                    # Return the directory of the first matching file
                    return os.path.dirname(file_path)
        
        raise ValueError("Could not find a calling module path outside of 'site-packages'.")           
    
    @classmethod
    def _update_gai_config(cls,name:str, gai_dict:dict, global_lib_config_path:str) -> "GaiGeneratorConfig":


        import glob
        
        here = cls._find_caller_module_path()
        # here = os.path.abspath(os.path.dirname(__file__))
        
        # traverse to the parent directory until we find the pyproject.toml
        
        while not os.path.exists(os.path.join(here, "pyproject.toml")):
            here = os.path.abspath(os.path.join(here, ".."))
            if here == "/":
                raise ValueError("GaiGeneratorConfig: pyproject.toml not found")
        
        # look for the file ./src/**/config/gai.yml
        namespace_dir = os.path.join(here, "src", "**", "config")
        config_path = os.path.join(namespace_dir, "gai.yml")
        config_files = glob.glob(config_path, recursive=True)
                        
        if not config_files:
            raise FileNotFoundError("GaiGeneratorConfig: gai.yml not found in the specified paths")

        # Read the first matching config file

        config_path = config_files[0]
        with open(config_path, 'r') as f:
            generator_dict = yaml.load(f, Loader=yaml.FullLoader)

        if not generator_dict.get("generators", None):
            raise ValueError(f"GaiGeneratorConfig: 'generators' key is required but not found in {config_path}.")
        
        if not generator_dict["generators"].get(name, None):
            raise ValueError(f"GaiGeneratorConfig: Generator Key {name} is not found in {config_path}.")
        
        # Update global config with the new generator config
        
        gai_dict["generators"][name] = generator_dict["generators"][name]
        
        with open(global_lib_config_path, 'w') as f:
            yaml.dump(gai_dict, f, sort_keys=False,indent=4)
                            
        logger.warning(f"GaiGeneratorConfig: Generator Config not found. name={name}. Config added to {global_lib_config_path}")            

        return generator_dict["generators"][name]

    @classmethod
    def _get_gai_config(cls, file_path:str) -> Dict:
        gai_dict = None
        try:
            with open(file_path, 'r') as f:
                gai_dict = yaml.load(f, Loader=yaml.FullLoader)
        except Exception as e:
            raise ValueError(f"GaiGeneratorConfig: Error loading client config from file: {e}")
        return gai_dict    
    
    @classmethod
    def _get_generator_config(
            cls,
            name: Optional[str] = None,
            config: Optional[dict] = None,
            file_path: Optional[str] = None    
        ) -> "GaiGeneratorConfig":
        if config:
            return cls(**config)
        if name:
            gai_dict = None
            try:
                app_dir=get_app_path()
                global_lib_config_path = os.path.join(app_dir, 'gai.yml')
                if file_path:
                    global_lib_config_path = file_path
                gai_dict = cls._get_gai_config(global_lib_config_path)
            except Exception as e:
                raise ValueError(f"GaiClientConfig: Error loading client config from file: {e}")

            generator_config = gai_dict["generators"].get(name, None)

            if not generator_config:
                
                # Self register the generator config in the gai config
                
                generator_config = cls._update_gai_config(
                    name=name, 
                    gai_dict=gai_dict, 
                    global_lib_config_path=global_lib_config_path)

            return cls(**generator_config)

        raise ValueError("GaiGeneratorConfig: Invalid arguments. Either 'name' or 'config' must be provided.")


### GaiConfig
    
class GaiConfig(BaseSettings):
    version: str
    gai_url: Optional[str] = None
    logging: Optional[LogConfig] = None
    clients: Optional[dict[str,GaiClientConfig] ] = None
    generators: Optional[dict[str,GaiGeneratorConfig] ] = None
    class Config:
        extra = "ignore"

    @classmethod
    def from_path(cls, file_path=None) -> "GaiConfig":
        """
        Class method to create an instance of GaiConfig from a YAML configuration file.

        Parameters:
            file_path (str, optional): Path to the configuration file. If not provided,
                                       the default path 'gai.yml' in the application directory is used.

        Returns:
            GaiConfig: An instance of GaiConfig populated with the configuration data.

        Usage example:
            config = GaiConfig.from_config()
        """
        app_dir=get_app_path()
        global_lib_config_path = os.path.join(app_dir, 'gai.yml')
        if file_path:
            global_lib_config_path = file_path
        with open(global_lib_config_path, 'r') as f:
            dict = yaml.load(f, Loader=yaml.FullLoader)
            gai_config = GaiConfig(**dict)
            return gai_config
        return config


from unittest.mock import patch, mock_open
from gai.lib.utils import get_rc, get_app_path
import os

@patch("builtins.open", new_callable=mock_open, read_data='{"app_dir": "~/.gai"}')
def test_get_rc(mock_file):
    rc=get_rc()
    
    # Ensure only ~/.gairc was opened
    mock_file.assert_called_once_with(os.path.expanduser("~/.gairc"), 'r')
    assert len(mock_file.call_args_list) == 1
    
    assert rc["app_dir"]=="~/.gai"

@patch("gai.lib.utils.get_rc", return_value={"app_dir": "/tmp"})
def test_get_app_path(mock_get_rc):
    
    # Should return the app_dir from the rc file
    assert get_app_path() == "/tmp"


import pytest
from unittest.mock import patch, mock_open, MagicMock,call
from gai.lib.config import GaiConfig

mock_gai_data= {
    "version": "1.0",
    "gai_url": "http://localhost:8080",
    "logging": {
        "level": "DEBUG",
        "format": "%(levelname)s - %(message)s"
    },
    "clients": {
        "ttt": {
            "type": "ttt",
            "engine": "ollama",
            "model": "llama3.1",
            "name": "llama3.1",
            "client_type": "ollama"
        }
    }
}

### GaiConfig should load from "~/.gai/gai.yml" by default

@patch("yaml.load", return_value=mock_gai_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.config.get_app_path", return_value="~/.gai")
def test_gaiconfig_from_path_default(mock_app_path, mock_file, mock_yaml_load):
    
    # Load GaiConfig from default path
    config = GaiConfig.from_path()
    
    # Ensure only ~/.gai/gai.yml was opened
    mock_file.assert_called_once_with("~/.gai/gai.yml", 'r')
    assert len(mock_file.call_args_list) == 1

### GaiConfig should load from custom path

@patch("yaml.load", return_value=mock_gai_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.config.get_app_path", return_value="~/.gai")
def test_gaiconfig_from_path_custom(mock_app_path, mock_file, mock_yaml_load):

    # Load Gaiconfig from a custom file path
    config = GaiConfig.from_path(file_path="/tmp/gai.yml")

    # Ensure only /tmp/gai.yml was opened
    mock_file.assert_called_once_with("/tmp/gai.yml", 'r')
    assert len(mock_file.call_args_list) == 1
    

mock_client_data= {
    "version": "1.0",
    "gai_url": "http://localhost:8080",
    "logging": {
        "level": "DEBUG",
        "format": "%(levelname)s - %(message)s"
    },
    "clients": {
        "ttt": {
            "type": "ttt",
            "engine": "ollama",
            "model": "llama3.1",
            "name": "llama3.1",
            "client_type": "ollama"
        }
    }
}

# GaiClientConfig should load from "~/.gai/gai.yml" by default
@patch("yaml.load", return_value=mock_client_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.config.get_app_path", return_value="~/.gai")
def test_clientconfig_from_path_default(mock_yaml_load, mock_file, mock_app_path):
    from gai.lib.config import GaiClientConfig
    
    # Load GaiConfig from default path
    config = GaiClientConfig.from_name("ttt")
    
    # Ensure only ~/.gai/gai.yml was opened
    mock_file.assert_called_once_with("~/.gai/gai.yml", 'r')
    assert len(mock_file.call_args_list) == 1

# GaiClientConfig should load from custom file path if provided
@patch("yaml.load", return_value=mock_client_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.config.get_app_path", return_value="~/.gai")
def test_clientconfig_from_custom_path(mock_yaml_load, mock_file, mock_app_path):
    from gai.lib.config import GaiClientConfig
    
    # Load GaiConfig from default path
    config = GaiClientConfig.from_name("ttt", file_path="/tmp/gai.yml")
    
    # Ensure only ~/.gai/gai.yml was opened
    mock_file.assert_called_once_with("/tmp/gai.yml", 'r')
    assert len(mock_file.call_args_list) == 1

# GaiClientConfig should not load from any file if the config is provided
@patch("yaml.load", return_value=mock_client_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.utils.get_app_path", return_value="~/.gai")
def test_clientconfig_from_dict(mock_yaml_load, mock_file, mock_app_path):
    from gai.lib.config import GaiClientConfig
    
    # Load GaiConfig from default path
    config = GaiClientConfig.from_dict({
        "type": "ttt",
        "engine": "ollama",
        "model": "llama3.1",
        "name": "llama3.1",
        "client_type": "ollama"
    })
    
    # Ensure only ~/.gai/gai.yml was opened
    assert len(mock_file.call_args_list) == 0


mock_generator_data= {
    "version": "1.0",
    "generators": {
        "ttt-exllamav2-dolphin": {
            "type": "ttt",
            "engine": "exllamav2",
            "model": "dolphin",
            "name": "ttt-exllamav2-dolphin",
            "module":{
                "name": "gai.ttt.server.gai_exllamav2",
                "class": "GaiExllamav2"
            }
        }
    }
}

# GaiGeneratorConfig should load from "~/.gai/gai.yml" by default
@patch("yaml.load", return_value=mock_generator_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.config.get_app_path", return_value="~/.gai")
def test_generatorconfig_from_path_default(mock_app_path, mock_file, mock_yaml_load):
    from gai.lib.config import GaiGeneratorConfig
    
    # Load GaiConfig from default path
    config = GaiGeneratorConfig.from_name("ttt-exllamav2-dolphin")
    
    # Ensure only ~/.gai/gai.yml was opened
    mock_file.assert_called_once_with("~/.gai/gai.yml", 'r')
    assert len(mock_file.call_args_list) == 1
    
    assert config.module.name == "gai.ttt.server.gai_exllamav2"

# GaiGeneratorConfig should load from custom file path if provided
@patch("yaml.load", return_value=mock_generator_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.config.get_app_path", return_value="~/.gai")
def test_generatorconfig_from_custom_path(mock_app_path, mock_file, mock_yaml_load):
    from gai.lib.config import GaiGeneratorConfig
    
    # Load GaiConfig from default path
    config = GaiGeneratorConfig.from_name("ttt-exllamav2-dolphin", file_path="/tmp/gai.yml")
    
    # Ensure only ~/.gai/gai.yml was opened
    mock_file.assert_called_once_with("/tmp/gai.yml", 'r')
    assert len(mock_file.call_args_list) == 1
    
    assert config.module.name == "gai.ttt.server.gai_exllamav2"

# GaiGeneratortConfig should not load from any file if the config is provided
@patch("yaml.load", return_value=mock_generator_data)
@patch("builtins.open", new_callable=mock_open, read_data="version: 1.0\ngai_url: http://localhost")
@patch("gai.lib.utils.get_app_path", return_value="~/.gai")
def test_generatorconfig_from_dict(mock_app_path, mock_file, mock_yaml_load):
    from gai.lib.config import GaiGeneratorConfig
    
    # Load GaiConfig from default path
    config = GaiGeneratorConfig.from_dict({
        "type": "ttt",
        "engine": "llamacpp",
        "model": "dolphin",
        "name": "ttt-llamacpp-dolphin",
        "module":{
            "name": "gai.ttt.server.gai_llamacpp",
            "class": "GaiLlamaCpp"
        }
    })
    
    assert config.module.name == "gai.ttt.server.gai_llamacpp"
    
    # Ensure only ~/.gai/gai.yml was opened
    assert len(mock_file.call_args_list) == 0





import subprocess
import os
import glob

def run_command(command):
    """Executes a shell command and returns the output and error if any."""
    try:
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        return result.stdout, None
    except subprocess.CalledProcessError as e:
        return e.stdout, e.stderr

def build_package():
    """Builds the package using poetry build and returns the path to the generated wheel file."""

    # Build the package
    stdout, stderr = run_command(["poetry", "build"])
    if stderr:
        print(f"Error during package build:{stderr}")
        return None
    print(stdout)

    # Find the wheel file in dist directory
    wheel_files = glob.glob('dist/*.whl')
    if wheel_files:
        # Assuming the latest built wheel file is the one we want (last modified)
        return max(wheel_files, key=os.path.getmtime)
    else:
        print("No wheel file found in dist directory.")
        return None

def install_wheel(wheel_path):
    """Installs a wheel file using pip."""
    if wheel_path:
        stdout, stderr = run_command(["pip", "install", wheel_path,"--force-reinstall"])
        if stderr:
            print(f"Error during wheel installation:{stderr}")
        else:
            print(f"Wheel installed successfully.\n{stdout}")
    else:
        print("Installation skipped as no wheel path was provided.")

def main():
    print("Starting the build process...")
    wheel_path = build_package()
    if wheel_path:
        print(f"[green]Wheel built at {wheel_path}. Starting installation...")
        os.system("which python")
        install_wheel(wheel_path)
    else:
        print("[red]Build failed or no wheel file was produced.")

if __name__ == "__main__":
    main()



import re
from typing import List

# This is used to compress a list into a smaller string to be passed as a single USER message to the prompt template.
def chat_list_to_string(messages):
    if type(messages) is str:
        return messages
    prompt=""        
    for message in messages:
        if prompt:
            prompt+="\n"
        content = message['content'].strip()
        role = message['role'].strip()
        if content:
            prompt += f"{role}: {content}"
        else:
            prompt += f"{role}:"
    return prompt

# This is useful for converting text dialog to chatgpt-style dialog
def chat_string_to_list(messages):
    # Split the messages into lines
    lines = messages.split('\n')

    # Prepare the result list
    result = []

    # Define roles
    roles = ['system', 'user', 'assistant']

    # Initialize current role and content
    current_role = None
    current_content = ''

    # Process each line
    for line in lines:
        # Check if the line starts with a role
        for role in roles:
            if line.lower().startswith(role + ':'):
                # If there is any content for the current role, add it to the result
                if current_role is not None and current_content.strip() != '':
                    result.append({'role': current_role, 'content': current_content.strip()})
                
                # Start a new role and content
                current_role = role
                current_content = line[len(role) + 1:].strip()
                break
        else:
            # If the line does not start with a role, add it to the current content
            current_content += ' ' + line.strip()

    # Add the last role and content to the result
    if current_role is not None:
        result.append({'role': current_role, 'content': current_content.strip()})

    return result

def chat_list_to_INST(input_list):
    # Initialize an empty string for the output
    output = "<s>\n\t[INST]\n"
    
    # if last message is an AI placeholder, remove it
    last_role = input_list[-1]["role"].lower()
    last_content = input_list[-1]["content"]
    if last_role != "system" and last_role != "user" and last_content == "":
        input_list.pop()

    # Loop through the list of dictionaries
    for item in input_list:
        # Check the role
        role = item["role"].lower()
        if role == "system":
            # Add the system message
            output += f"\t\t<<SYS>>\n\t\t\t{item['content']}\n\t\t<</SYS>>\n"
        elif role == "user":
            # Add the user message
            output += f"\t\t{item['content']}\n"
            output += "\t[/INST]\n\n\t"
        else:
            # Add the AI message
            output += f"{item['content']}\n\n"
            # AI message marks the end of 1 turn
            output += "</s>\n"
            # Add the beginning of next turn
            output += "<s>\n\t[INST]\n"
   
    return output

def INST_output_to_output(output_string):
    # The rfind method returns the last index where the substring is found
    last_index = output_string.rfind('[/INST]\n\n\t')

    # Add the length of '[/INST]\n\n\t' to get the start of the desired substring
    start_of_substring = last_index + len('[/INST]\n\n\t')

    # Extract the substring from start_of_substring till the end of the string
    result = output_string[start_of_substring:]

    return result

def ASSISTANT_output_to_output(output_string):
    return re.split('\n.+:',output_string)[-1].strip()




import os, re, time
import json
from gai.lib import constants

# Get JSON FROM ~/.gairc
def get_rc():
    if (not os.path.exists(os.path.expanduser(constants.GAIRC))):
        raise Exception(f"Config file {constants.GAIRC} not found. Please run 'gai init' to initialize the configuration.")
    with open(os.path.expanduser(constants.GAIRC), 'r') as f:
        return json.load(f)

# Get "app_dir" from ~/.gairc
def get_app_path():
    rc = get_rc()
    app_dir=os.path.abspath(os.path.expanduser(rc["app_dir"]))
    return app_dir




from pathlib import Path
from rich.console import Console
import shutil
import json,os
import importlib.resources as pkg_resources

PACKAGED_DATA_PATH = pkg_resources.path('gai.scripts.data', '')
PACKAGED_CONFIG_PATH = pkg_resources.path('gai.scripts.data', 'gai.yml')

def init(force=False):
    console=Console()

    # Initialise config

    if not force and (Path("~/.gairc").expanduser().exists() or Path("~/.gai").expanduser().exists()):
        console.print("[red]Already exists[/]")
        return
    
    # app_dir doesn't exist
    if not Path("~/.gai").expanduser().exists():
        Path("~/.gai").expanduser().mkdir()

    # models directory doesn't exist
    if not Path("~/.gai/models").expanduser().exists():
        Path("~/.gai/models").expanduser().mkdir()

    # source_persona_path = PACKAGED_DATA_PATH / "persona"
    
    # if not Path("~/.gai/persona/nodes").expanduser().exists():
    #     Path("~/.gai/persona/nodes").expanduser().mkdir(parents=True)
        
    #     # cp -rp gai-data/src/gai/data/persona/* ~/.gai/persona/nodes
    #     for item in Path(source_persona_path).glob("*"):
    #         if item.is_dir():
    #             shutil.copytree(item, Path("~/.gai/persona/nodes").expanduser() / item.name)
    #         else:
    #             shutil.copy(item, Path("~/.gai/persona/nodes").expanduser() / item.name)

    # Force create .gairc
    with open(Path("~/.gairc").expanduser(), "w") as f:
        f.write(json.dumps({
            "app_dir":"~/.gai"
        }))

    # Finally
    if Path("~/.gairc").expanduser().exists() and Path("~/.gai").expanduser().exists():
        console.print("[green]Initialized[/]")
    
    # Copy all files in PACKAGED_DATA_PATH to ~/.gai
    DESTINATION = Path("~/.gai").expanduser()
    ignore_patterns = shutil.ignore_patterns('__pycache__', '__init__.py')    
    for item in PACKAGED_DATA_PATH.glob("*"):
        if item.is_dir():
            shutil.copytree(
                item,
                DESTINATION / item.name,
                ignore=ignore_patterns
            )
        else:
            # Skip __init__.py files in the top-level
            if item.name != '__init__.py':
                shutil.copy(item, DESTINATION / item.name)        
        
    
if __name__=="__main__":
    init(True)


import os
from rich.console import Console
console=Console()
from gai.scripts._docker_utils import _docker_push

def docker_push(pyproject_path, 
                repo_name,
                image_name
                ):
    _docker_push(
        pyproject_path=pyproject_path,
        repo_name=repo_name,
        image_name=image_name)
    
